#!/usr/bin/env python3

import sys
import matplotlib
matplotlib.use("Agg")
print('[A] Loading ' + sys.argv[0] + '.\n')
import os # for reading files
import networkx as nx # for visualisation
from copy import deepcopy
from nltk.corpus import stopwords
import numpy as np # for calculations
import re
import spacy # for nlp
from multiprocessing import Pool
import random
import matplotlib.pyplot as plt
import config

nlp = spacy.load('en') # standard english nlp

def frequencies(target_string, search_result_list):
    """Counts occurrences of nodes and cooccurrences.
    
    Iterates over the corpus (and snippets provided with the task) line by line 
    and counts every token and tuple of tokens within a line (context). These 
    tokens is filtered by stop words, pos tags and context length.
    
    Args:
        target_string: contexts are selected if they contain this string. For
            further processing this string is removed from the contexts.
        search_result_list: List of titles and snippets provided with the task.
        
    Returns:
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    """

    corpus_path = config.corpus
    max_node_count = config.max_nodes
    max_edge_count = config.max_edges
    
    bracketed_target_string = '('+target_string+')'
    
    # Remove unnecessary tokens from snippets 
    _search_result_list = list()
    for r in search_result_list:
        r = r.replace('<b>', '')
        r = r.replace('</b>', '')
        r = r.replace(r'\\', '')
        r = r.strip()
        _search_result_list.append(r)
    
    #initialises frequencies with counts from results
    node_freq_dict, edge_freq_dict = process_file(_search_result_list,
                                                  target_string,
                                                  dict(),
                                                  dict()) 
    
    #names of corpus files
    corpus_file_path_list = [corpus_path + f for f in os.listdir(corpus_path)]
    corpus_size = len(corpus_file_path_list)
    
    processed_file_count = 0
    for corpus_file_path in corpus_file_path_list:
        
        node_count = len(node_freq_dict)
        edge_count = len(edge_freq_dict)
        
        #prints update after every 11th of the corpus is parsed
        if processed_file_count % int(corpus_size/11) == 0: 
            
            file_ratio = processed_file_count / corpus_size
            max_node_ratio = node_count / max_node_count
            max_edge_ratio = edge_count / max_edge_count
            
            ratios = [file_ratio, max_node_ratio, max_edge_ratio]
            
            #uses the ratio closest to 100%.
            highest_ratio = int((max(ratios))*100)
            
            print('[a] ~{:02d}%\tNodes: {}\tEdges: {}\t{}.'.format(highest_ratio,
                                                                   node_count,
                                                                   edge_count,
                                                                   bracketed_target_string))
        
        #checks maximum node values
        if node_count > max_node_count:
            print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                               edge_count,
                                                               bracketed_target_string))
            return node_freq_dict, edge_freq_dict
        
        if edge_count > max_edge_count:
            print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                               edge_count,
                                                               bracketed_target_string))
            return node_freq_dict, edge_freq_dict
        
        with open(corpus_file_path, 'r') as corpus_file:
            
            node_freq_dict, edge_freq_dict = process_file(corpus_file,
                                                          target_string,
                                                          node_freq_dict,
                                                          edge_freq_dict)
        
        processed_file_count += 1
    
    print('[a] 100%\tNodes: {}\tEdges: {}\t{}.'.format(node_count,
                                                       edge_count,
                                                       bracketed_target_string))
    
    return node_freq_dict, edge_freq_dict


def process_file(context_list, target_string, node_freq_dict, edge_freq_dict):
    """Updates the counts of nodes and edges for a given document and target.
    
    Ammends the input dictionaries with counts from each context withing the
    list of contexts. Furthermore filters out small contexts and tokens from
    the stopword list or with wrong pos tags.
    
    Args:
        context_list: List of contexts (lines, paragraphs) that are to be
            considered for updating the counting dictionaries.
        target_string: Target string for filtering out every context that does 
            not contain it.
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    Returns:
        node_freq_dict: Updated version of the input node dict.
        edge_freq_dict: Updated version of the input edge dict.
    """
    
    spaced_target_string = target_string.replace('_', ' ')
    
    stopword_list = set(stopwords.words('english') + config.stop_words)
    allowed_tag_list = config.allowed_tags
    min_context_size = config.min_context_size
        
    try:
        
        for context in context_list:
            
            context = context.lower()
            if spaced_target_string in context: #greedy pre selection, not perfect
                
                token_set = set() #set of node candidates
                
                #This replacement allows target to be treated as single entity.
                context = context.replace(spaced_target_string, target_string)
                processed_context = nlp(context)
                
                if target_string in [token.text for token in processed_context]:
                    
                    for token in processed_context:
                        
                        #doesn't add target word to nodes
                        if token.text == target_string:
                            pass
                        
                        #doesn't add stop words to nodes
                        elif token.text in stopword_list:
                            pass
                        
                        #only adds tokens with allowed tags to nodes
                        elif token.tag_ in allowed_tag_list:
                            token_set.add(token.text)
                            
                    context_size = len(token_set)
                    
                    if context_size >= min_context_size:
                        for token in token_set:
                            
                            if token in node_freq_dict:
                                node_freq_dict[token] += 1
                            else:
                                node_freq_dict[token] = 1
                        
                        #set of possible edges
                        for edge in {(x,y) for x in token_set for y in token_set if x < y}:
                            
                            if edge in edge_freq_dict:
                                edge_freq_dict[edge] += 1
                            else:
                                edge_freq_dict[edge] = 1
    
    #if a file is corrupted (can't always be catched with if-else)
    except UnicodeDecodeError:
        
        pass            
    
    return node_freq_dict, edge_freq_dict

#build graph from frequency dictionaries
def build_graph(node_freq_dict, edge_freq_dict):
    """Builds undirected weighted graph from dictionaries.
    
    Creates graph and appends every edge and node in the parameter dictionaries,
    given they occur frequently enough. For every edge a weight is calculated.
    
    Args:
        node_freq_dict: Dictionary of occurrences of every eligible token
            within every context the target occurs in.
        edge_freq_dict: Dictionary of occurrences of every eligible tuple of
            tokens within every context the target occurs in.
    
    Returns:
        cooccurence_graph: Filtered undirected dice weighted small word 
            cooccurence graph for a given target entity.
    """
    
    min_node_freq = config.min_node_freq
    min_edge_freq = config.min_edge_freq
    max_weight = config.max_weight
    
    cooccurence_graph = nx.Graph()
    
    #node : node frequency
    for node, frequency in node_freq_dict.items():
        
        if frequency >= min_node_freq:
            cooccurence_graph.add_node(node)
            
    #edge : edge frequency
    for node_tuple, frequency in edge_freq_dict.items():
        
        if frequency < min_edge_freq:
            
            continue
        
        elif node_tuple[0] not in cooccurence_graph.nodes:
            
            continue
        
        elif node_tuple[1] not in cooccurence_graph.nodes:
            
            continue
        
        else:
            
            cooccurrence_frequency = edge_freq_dict[node_tuple]
            node0_frequency = node_freq_dict[node_tuple[0]]
            node1_frequency = node_freq_dict[node_tuple[1]]
            
            prob_0 = cooccurrence_frequency / node0_frequency
            prob_1 = cooccurrence_frequency / node1_frequency
            
            #best_weight = 1 - max(prob_0, prob_1)
            dice_weight = 1 - ((prob_0 + prob_1) / 2)
            
            if dice_weight <= max_weight:
                
                cooccurence_graph.add_edge(*node_tuple, weight=dice_weight)
            
            else:
                
                pass
    
    return cooccurence_graph


#Identifies senses by choosing nodes with high degrees
def root_hubs(graph, edge_freq_dict, min_neighbors=4, theshold=0.8):
    
    min_neighbors = config.min_neighbors
    threshold = config.threshold
    
    G = deepcopy(graph)
    V = sorted(G.nodes, key=lambda key: G.degree[key], reverse=True) # sorts according to degree
    H = list() #output list
    
    while V:
        
        v = V[0] #best hub candidate
        
        if G.degree[v] >= min_neighbors:
        
            mfn = sorted(G.adj[v], key=lambda key: edge_freq_dict[v,key] if v < key else edge_freq_dict[key, v], reverse=True)[:min_neighbors] #most frequent neighbors
            
            if np.mean([G.edges[v,n]['weight'] for n in mfn]) < theshold: #if the median weight of the most frequent neighbors is under threshold
                
                H.append(v)
            
                #removes neighbors of new hub as hub candidates
                for nbr in deepcopy(G).adj[v]:
                
                    G.remove_node(nbr)
                
            #removes hub candidate
            G.remove_node(v)
            
            #reorderd potential hubs after deletions
            V = sorted(G.nodes, key=lambda key: G.degree[key], reverse=True)
        
        else:
        
            return H
    
    return H


#Components algorithm from Véronis (2004), converts graph for target into a MST
def components(graph, hubs, target_string):
    
    G = deepcopy(graph)
    H = hubs #root hubs
    t = target_string
    
    #G.add_node(t)
    #for h in H:
        #G.add_edge(t,h,weight=0)
        
    T = nx.minimum_spanning_tree(G)
    
    #removes singletons
    for node in deepcopy(T).nodes:
        if len(T.adj[node]) == 0:
            T.remove_node(node)
    
    return T


#Calculates score for a given path in a minimum spanning tree
def score(graph, from_node, to_node):
    
    #if correct tree
    if nx.has_path(graph, from_node, to_node):
                
        # calculates shortest path (approximation for path with lowest total weight)
        path = nx.shortest_path(graph, from_node, to_node, 'weight')
        total_weight = 0
    
        #adds weights of every sub-path
        for i in range(1, len(path)):
            sub_from, sub_to = path[i-1], path[i]
            total_weight += graph[sub_from][sub_to]['weight']
    
        #the further the path, the lower the score
        return 1/(1+total_weight)
        
    else:
        
        return 0


# Basically Word Sense Disambiguation, matches context to sense
def disambiguate(mst, hubs, contexts, target_string):
    
    target_string = target_string.replace('_', ' ')
    T = mst #minimum spanning tree
    H = hubs #root hubs
    C = [c.lower().strip().replace(target_string, '') for c in contexts] #cleaned up contexts
    
    score_dict = dict() #memoisation for scores
    mapping_dict = {topic:[] for topic in range(1,len(H)+1)} #output of function
    
    #if no sense is found for a target word, we should assume that there only is one sense
    if len(H) == 0: 
            
        return {0:[i for i in range(1, len(C)+1)]}
    
    idx = 0
    
    for c in C:
        
        idx += 1 #index based on position in list
    
        doc = nlp(c) #parsed context
        texts = [tok.text for tok in doc] #tokens
        
        scores = np.zeros(len(H)) #initialise with zeros for every sense
        
        for text in texts:
            
            if text in T.nodes: #if word wasn't filtered out
                
                new_scores = list() #scores to be added to total scores
                
                for h in H: #for each hub
                    
                    if (text, h) in score_dict: #memoisation
                        
                        new_scores.append(score_dict[(text,h)])
                    
                    else:
                        
                        new_score = score(T, text, h)
                        new_scores.append(new_score)
                        score_dict[(text,h)] = new_score #memoisation
                    
                scores = scores + np.array(new_scores)
            
            else:
            
                pass
        
        #if the disambiguator could not detect a sense, it should return a singleton, ie. nothing
        if np.max(scores) == 0:
            
            pass
            
        else:
                
            #applies sense with the highest score to context
            max_score = np.max(scores)
            argmax_score = np.argmax(scores)
            
            #clusters begin at 1
            mapping_dict[argmax_score + 1].append(idx)

    return mapping_dict


# our main function, here the main stepps for word sense induction are called
def word_sense_induction(topic_id, topic_name, results):
    
    #buffer for useful information
    out_buffer = '\n'
    
    #path for output(directory)
    output_path = './test/'#config.output
            
    #removes trailing new_lines
    old_target_string = topic_name.strip() #original target
    
    if old_target_string.strip() in [f.replace('.absinth', '') for f in os.listdir(config.output)]:
        return None
    
    out_buffer += ("[A] Word sense induction for '"+old_target_string+"':\n")
    
    #in topics longer than two words, the leading 'the' can generally be removed without changing the sense
    if old_target_string[:4] == 'the_' and old_target_string.count('_') >= 2:
        
        target_string = old_target_string[4:]
        
    else:
        
        target_string = old_target_string
    
    #writes headline for output files
    f = open(output_path+target_string+'.absinth', 'w')
    f.write('subTopicID\tresultID\n')
    
    #counts occurences of single words, as well as cooccurrences, saves it in dictionary
    print('[a]', 'Counting nodes and edges.\t('+old_target_string+')')
    node_freq_dict, edge_freq_dict = frequencies(target_string, results[topic_id])
    
    #builds graph from these dictionaries, also applies multiple filters
    print('[a]', 'Building graph.\t('+old_target_string+')')
    G = build_graph(node_freq_dict, edge_freq_dict)
    out_buffer += '[A] Nodes: {}\tEdges: {}\n'.format(str(len(G.nodes)), str(len(G.edges)))
    
    #finds root hubs (senses) within the graph + more filters for these
    print('[a]', 'Collecting root hubs.\t('+old_target_string+')')
    H = root_hubs(G, edge_freq_dict)
    out_buffer += '[A] Root hubs:\n'
    
    #adds sense inventory to buffer with some common neighbors for context
    i = 1 #sense index
    for h in H:
        
        mfn = sorted(G.adj[h], key=lambda x: edge_freq_dict[h,x] if h < x else edge_freq_dict[x, h], reverse=True)[:6]
        out_buffer += (' {}. {}: {}\n'.format(i, h, ', '.join(mfn)))
        i += 1
    
    #performs minimum_spanning_tree algorithm on graph
    print('[a]', 'Building minimum spanning tree.\t('+old_target_string+')')
    T = components(G, H, target_string)

    #matches senses to clusters
    print('[a]', 'Disambiguating results.\t('+old_target_string+')')
    D = disambiguate(T, H, results[topic_id], target_string)
    
    out_buffer += ('[A] Mapping: \n')
    for cluster,results in D.items():
        out_buffer += (' {}. : {}\n'.format(cluster, ', '.join([str(r) for r in results])))
    
    #prints buffer
    print('[a]', 'Writing to file.\t('+old_target_string+')')
    print(out_buffer)
    
    #writes clustering to file
    for cluster,results in D.items():
        for result in results:
            f.write(topic_id+'.'+str(cluster)+'\t'+topic_id+'.'+str(result)+'\n')
        
    f.close()
    
    
def read_dataset(data_path):
    
    # results.txt includes the queries for a given target word
    results = dict()
    
    with open(data_path+'results.txt', 'r') as results_file:
        
        for line in results_file.readlines()[1:]:
            
            l = line.split('\t')
            id1, _ = l[0].split('.') #the second part of the id is ignored, as it is identical to the list index
            
            if id1 not in results:
                results[id1]=list()
                
            results[id1].append(" ".join(l[2:])) # here I join title and snippet, the URL is ignored
            
    
    # topics.txt is a list of target words
    topics = dict()
    
    with open(data_path+'topics.txt', 'r') as topics_file:
        
        for line in topics_file.readlines()[1:]:
            
            l = line.split('\t')
            topics[l[0]] = l[1]
    
    return results, topics


def main():
    
    # If absinth.py is run in test environment
    if '-t' in sys.argv:
        data_path = config.test
    else:
        data_path = config.dataset
        
    results, topics = read_dataset(data_path)
    
    with Pool(2) as pool:
        parameter_list = [(topic_id, topic_name, results)
                          for topic_id,topic_name in topics.items()]
        pool.starmap(word_sense_induction, parameter_list)
        
    #for topic_id,topic_name in topics.items():
       #word_sense_induction(topic_id,topic_name, results)


if __name__ == '__main__':
    main()
